/*
 * Copyright 2021 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafkarest.entities;

import com.google.auto.value.AutoValue;
import io.confluent.kafka.schemaregistry.ParsedSchema;

@AutoValue
public abstract class RegisteredSchema {

  RegisteredSchema() {
  }

  public abstract String getSubject();

  public abstract int getSchemaId();

  public abstract int getSchemaVersion();

  public abstract ParsedSchema getSchema();

  public final EmbeddedFormat getFormat() {
    return EmbeddedFormat.forSchemaType(getSchema().schemaType());
  }

  public static RegisteredSchema create(
      String subject, int schemaId, int schemaVersion, ParsedSchema schema) {
    return new AutoValue_RegisteredSchema(subject, schemaId, schemaVersion, schema);
  }
}
